char *
truepath(const char *path, char resolved_path[MAXPATHLEN])
{
	unsigned num_symlinks = 0;
	int saved_errno = errno;

	char left[MAXPATHLEN];
	size_t left_len, resolved_len;

	if (path[0] == '/') {
		resolved_path[0] = '/';
		resolved_path[1] = '\0';
		if (path[1] == '\0')
			return resolved_path;
		resolved_len = 1;
		left_len = strlcpy(left, path + 1, MAXPATHLEN);
	} else {
		if (getcwd(resolved_path, MAXPATHLEN) == NULL) {
			strlcpy(resolved_path, ".", MAXPATHLEN);
			return NULL;
		}
		resolved_len = strlen(resolved_path);
		left_len = strlcpy(left, path, MAXPATHLEN);
	}
	if (left_len >= MAXPATHLEN || resolved_len >= MAXPATHLEN) {
		errno = ENAMETOOLONG;
		return NULL;
	}

	while (left_len > 0) {
		struct stat st;
		char next_token[MAXPATHLEN];
		char *p;
		char *s = (p = strchr(left, '/')) ? p : left + left_len;

		memmove(next_token, left, s - left);
		left_len -= s - left;
		if (p != NULL)
			memmove(left, s + 1, left_len + 1);

		next_token[s - left] = '\0';
		if (resolved_path[resolved_len - 1] != '/') {
			if (resolved_len + 1 >= MAXPATHLEN) {
				errno = ENAMETOOLONG;
				return NULL;
			}

			resolved_path[resolved_len++] = '/';
			resolved_path[resolved_len] = '\0';
		}

		if (next_token[0] == '\0')
			continue;
		else if (!strcmp(next_token, "."))
			continue;
		else if (!strcmp(next_token, "..")) {
			if (resolved_len > 1) {
				char *q;

				/* trailing slash */
				resolved_path[resolved_len - 1] = '\0';

				q = strrchr(resolved_path, '/');
				*q = '\0';
				resolved_len = q - resolved_path;
			}
			continue;
		}

		/* filename */
		resolved_len = strlcat(resolved_path, next_token, MAXPATHLEN);
		if (resolved_len >= MAXPATHLEN) {
			errno = ENAMETOOLONG;
			return NULL;
		}

		if (lstat(resolved_path, &st) < 0) {
			if (errno == ENOENT && p == NULL) {
				errno = saved_errno;
				return resolved_path;
			}

			return NULL;
		}

		if ((st.st_mode & S_IFLNK) == S_IFLNK) {
			char symlink[MAXPATHLEN];
			int slen;

			if (num_symlinks++ > MAXSYMLINKS) {
				errno = ELOOP;
				return NULL;
			}
			slen = readlink(resolved_path, symlink, MAXPATHLEN);
			if (slen < 0)
				return NULL;
			symlink[slen] = '\0';

			if (symlink[0] == '/') {
				/* absolute link */
				resolved_path[1] = 0;
				resolved_len = 1;
			} else if (resolved_len > 1) {
				char *q;

				/* trailing slash */
				resolved_path[resolved_len - 1] = '\0';

				q = strrchr(resolved_path, '/');
				*q = '\0';
				resolved_len = q - resolved_path;
			}

			if (symlink[slen - 1] != '/' && p != NULL) {
				if (slen >= MAXPATHLEN) {
					errno = ENAMETOOLONG;
					return NULL;
				}

				symlink[slen] = '/';
				symlink[slen + 1] = 0;
			}
			if (p != NULL)
				left_len = strlcat(symlink, left, MAXPATHLEN);
			if (left_len > MAXPATHLEN) {
				errno = ENAMETOOLONG;
				return NULL;
			}
			left_len = strlcpy(left, symlink, MAXPATHLEN);
		}
	}

	if (resolved_len > 1 && resolved_path[resolved_len - 1] == '/')
		resolved_path[resolved_len - 1] = '\0';

	return resolved_path;
}
